# -*- Perl -*-

package SGML::DTDParse;

$VERSION = "0.97";

use strict;
use Env 'SGML_CATALOG_FILES';
use Text::DelimMatch;

require 5.000;
require Carp;

{
    package SGML::DTDParse::DTD_Element;

    # Expose these routines in this package too...
    use Text::DelimMatch;

    sub new {
	my($type, $dtd, %assoc) = @_;
	my($class) = ref($type) || $type;
	my($self) = {};
	my($key, $val);

	$self->{'DTD'} = $dtd;

	while (($key, $val) = each %assoc) {
	    $self->{$key} = $val;
	}

	bless $self, $class;
    }

    sub load {
	my($self) = shift;
	local(*FILE) = shift;
	my(@fields, $key, $value);
	my($fs) = "\034";
	local($_);
	
	chop($_ = scalar(<FILE>));
	@fields = split(/$fs/, $_);

	foreach $_ (@fields) {
	    ($key, $value) = /^(\S+?)=(.*)$/;
	    $self->{$key} = $value if defined($key) && defined($value);
	}
    }

    sub store {
	my($self) = shift;
	local(*FILE) = shift;
	my(%keys) = ();
	my($fs) = "\034";
	my($key, $value);
	local($_) = "";

	foreach $key (keys %{$self}) {
	    $keys{$key} = 1;
	}

	delete $keys{'DTD'};

	foreach $key (keys %keys) {
	    $value = $self->{$key};
	    $value = join(" ", sort values %{$value})
		if ref $value eq 'HASH';
	    $_ .= $fs if $_ ne "";
	    $_ .= "$key=$value";
	}

	print FILE ref $self, "\n";
	print FILE $_, "\n";
    }

    sub name {
	my($self) = @_;
	return $self->{'NAME'};
    }

    sub content {
	my($self) = @_;
	return $self->{'CONT'};
    }

    sub content_model_paren_match {
	my($self, $cm) = @_;
	my($mc) = new Text::DelimMatch '\(', '\)[\?\+\*\,]*';
	my($pre, $match, $rest) = $mc->match($cm);
	return ("$pre$match", $rest);
    }

    sub format_content_model {
	my($self) = shift;
	local($_) = shift;
	my($indent) = shift;
	my(@lines) = ();
	my($match, $rest, @rem);
	my($dtd) = $self->{'DTD'};
	my(@cm) = ();

	if ($dtd->{'EXPAND'}) {
	    $_ = $dtd->expand_entities($_);
	}

	$indent = 2 if !defined($indent);

	if (/\s+/ && !(/\|/ || /\%/)) {
	    # it doesn't really look like a content model to me...
	    push (@cm, " " x $indent . $_);
	    return @cm;
	}

	s/\s+//g;
	($match, $rest) = $self->content_model_paren_match($_);
	if (($match =~ /^\((.*)\)$/) && $rest eq "") {
	    $_ = $1;
	}

	@rem = ($_);

	while (@rem) {
	    $_ = shift @rem;

	    while (/\(.*$/) {
		my($pre) = $`;
		$_ = $&;

		push (@lines, $pre) if $pre ne "";

		(($match, $rest) = $self->content_model_paren_match($_));

		push (@lines, $match);

		$_ = $rest;
	    }

	    push (@lines, $_);
	}

	while (@lines) {
	    $_ = shift @lines;

	    next if /^\s*$/;

	    if (/^\((\(\(.*)\)([\?\+\*\,]*)$/) {
		my($body, $suffix) = ($1, $2);
		push (@cm, " " x $indent . "(");
		push (@cm, $self->format_content_model ($body, $indent+2));
		push (@cm, " " x $indent . ")$suffix");
	    } else {
		s/^\s*(.*?)\s*$/$1/g;
		
		if ((length($_) == 1) && @lines) {
		    $lines[0] = $_ . $lines[0];
		    next;
		} else {
		    my(@l2) = ($_);
		    
		    if (length($_) > 50) {
			@l2 = ();
			while (/^(.{1,50})([\|\,].*)$/) {
			    push (@l2, $1);
			    $_ = $2;
#			    print STDERR "->$1\n";
#			    print STDERR "=>$2\n";
			}
			push (@l2, $_);
		    }
	    
		    foreach $_ (@l2) {
			s/,/, /g;
			s/\|/ \| /g;
			s/([^\s])\%/$1 \%/g;
			s/^\s*(.*?)\s*$/$1/g; # in case we added it back on...
			push (@cm, " " x $indent . $_);
		    }
		}
	    }
	}

	return @cm;
    }
}

{
    package SGML::DTDParse::ELEMENT;
    @SGML::DTDParse::ELEMENT::ISA = qw (SGML::DTDParse::DTD_Element);

    sub parse {
	my($type, $dtd, $element) = @_;
	my($class) = ref($type) || $type;
	my($self) = {};
	my($name, $smin, $emin, $content, $incl, $excl);
	my($excontent);

	$self->{'DTD'}  = $dtd;
	$self->{'NAME'} = "";
	$self->{'MIN'}  = "";
	$self->{'INCL'} = "";
	$self->{'EXCL'} = "";
	$self->{'CONT'} = "";

	if ($element =~ /<!ELEMENT\s+(.*?)\s+([-o])\s+([-o])\s+(.*?)\s*>\s*$/i) {
	    $name = $1;
	    $smin = $2;
	    $emin = $3;
	    $content = $4;
	} elsif ($element =~ /<!ELEMENT\s+(.*?)\s+(.*?)\s*>\s*$/i) {
	    $name = $1;
	    $smin = "-";
	    $emin = "-";
	    $content = $2;
	} else {
	    print "Element? $element\n";
	    return bless $self, $class;
	}
	$name = $1;
	$smin = $2;
	$emin = $3;
	$content = $4;

	# expand the content, in case inclusions or exclusions are hidden
	# in parameter entities...
	$excontent = $dtd->expand_entities ($content);

	$incl = "";
	$excl = "";

	while ($excontent =~ /\+\((.*?)\)/) {
	    $incl .= " " if $incl ne "";
	    $incl .= $1;
	    $excontent = "$`  $'";
	}

	while ($excontent =~ /\-\((.*?)\)/) {
	    $excl .= " " if $excl ne "";
	    $excl .= $1;
	    $excontent = "$`  $'";
	}

	$self->{'NAME'} = $name;
	$self->{'MIN'} = "$smin $emin";
	$self->{'INCL'} = $incl;
	$self->{'EXCL'} = $excl;
	$self->{'CONT'} = $content;
	$self->{'CONTEX'} = $excontent;
	
	bless $self, $class;
    }

    sub parents {
	no strict;
	my($self) = @_;
	my(@p) = split(/\s+/, $self->{'PARENTS'});
	my(@ip) = split(/\s+/, $self->{'IPARENTS'});
	my(@ps) = sort { uc($a) cmp uc($b) } (@p, @ip);

	return join (" ", @ps);
    }

    sub children {
	no strict;
	my($self) = @_;
	my(@c) = split(/\s+/, $self->{'CHILDREN'});
	my(@ic) = split(/\s+/, $self->{'ICHILDREN'});
	my(@cs) = sort { uc($a) cmp uc($b) } (@c, @ic);

	return join (" ", @cs);
    }
}

{
    package SGML::DTDParse::ENTITY;
    @SGML::DTDParse::ENTITY::ISA = qw (SGML::DTDParse::DTD_Element);

    %SGML::DTDParse::ENTITY::EFILES = ();

    sub parse {
	my($type, $dtd, $entity) = @_;
	my($class) = ref($type) || $type;
	my($self) = {};

	$self->{'DTD'}  = $dtd;
	$self->{'NAME'} = "";
	$self->{'TYPE'} = "";
	$self->{'CONT'} = "";
	$self->{'PENT'} = 0;
	$self->{'FILE'} = 0;

	if ($entity =~ /<!ENTITY\s+\%\s*(\S+)\s+([\"\'])(.*?)\2\s*>/is) {
	    my($name, $cont) = ($1, $3);
	    $self->{'NAME'} = $name;
	    $self->{'CONT'} = $cont;
	    $self->{'PENT'} = 1;
	} elsif ($entity =~ /<!ENTITY\s+\%\s*(\S+)\s+(\S+)\s+([\"\'])(.*?)\3\s*>/is) {
	    my($name, $type, $cont) = ($1, $2, $4);
	    $type = uc($type);

	    warn "Unexpected type: $entity\n" 
		if ($type ne 'SYSTEM') && ($type ne 'PUBLIC');

	    $self->{'NAME'} = $name;
	    $self->{'TYPE'} = $type;
	    $self->{'CONT'} = $cont;
	    $self->{'PENT'} = 1;
	    $self->{'FILE'} = 1;
	} elsif ($entity =~ /<!ENTITY\s+(\S+)\s+(\S+)\s+([\"\'])(.*?)\3\s*>/is) {
	    my($name, $type, $cont) = ($1, $2, $4);
	    $self->{'NAME'} = $name;
	    $self->{'TYPE'} = $type;
	    $self->{'CONT'} = $cont;
	} else {
	    print "Entity? $entity\n";
	}

	bless $self, $class;
    }

    sub type {
	my($self, $value) = @_;

	return ($self->{'TYPE'} eq $value) if defined($value);
	return $self->{'TYPE'};
    }

    sub content {
	my($self) = @_;

	if (($self->{'TYPE'} eq 'PUBLIC')
	    || ($self->{'TYPE'} eq 'SYSTEM')) {
	    return $self->load_content();
	} else {
	    return $self->{'CONT'};
	}
    }

    sub content_entity {
	my($self) = @_;

	# An entity is a content entity if it is not a file parameter 
	# entity and does not have the value "INCLUDE" or "IGNORE"

	return ($self->{'FILE'} 
		|| ($self->{'PENT'} 
		    && ($self->{'CONT'} =~ /^ignore|include$/i))) ? 0 : 1;
    }

    sub load_content {
	my($self) = @_;
	local($_);

	return $SGML::DTDParse::ENTITY::EFILES{$self->{'NAME'}}
	    if defined($SGML::DTDParse::ENTITY::EFILES{$self->{'NAME'}});

	if ($self->{'TYPE'} eq 'PUBLIC') {
	    $_ = $self->load_public_content();
	} elsif ($self->{'TYPE'} eq 'SYSTEM') {
	    $_ = $self->load_system_content();
	} else {
	    warn "Not a file identifier? $self->{NAME}.\n";
	    $_ = "";
	}

#	s/\n/ /g;        # get rid of newlines
#	s/\s+/ /g;       # turn multiple spaces into a single space
	s/<!--.*?-->//sg; # discard all the comments
	s/--.*?--//sg;    # discard all the comments (in other elements)

	$SGML::DTDParse::ENTITY::EFILES{$self->{'NAME'}} = $_;

	return $_;
    }

    sub load_public_content {
	my($self) = @_;
	my($dtd) = $self->{'DTD'};
	my($file) = $dtd->{'CATALOG'}->{$self->{'CONT'}};
	local (*F, $_);

	# If we can't find "foo//bar..." look for "-//foo//bar...";
	$file = $dtd->{'CATALOG'}->{"-//" . $self->{'CONT'}}
	    if !defined($file);

	if (!defined($file)) {
	    warn "No pubid map in catalogs: $self->{CONT}.\n";
	    return "";
	}
	
	if (!open (F, $file)) {
	    warn "Cannot read $file for $self->{CONT}.\n";
	    return "";
	}
	    
	read (F, $_, -s $file);
	close (F);

	print "PUBLIC $self->{NAME}: $file.\n"
	    if $dtd->{'VERBOSE'};

	return $_;
    }

    sub load_system_content {
	my($self) = @_;
	my($dtd) = $self->{'DTD'};
	my($file) = $self->{'CONT'};
	local (*F, $_);

	if (!open (F, $file)) {
	    warn "Cannot read $file.\n";
	    return "";
	}
	    
	read (F, $_, -s $file);
	close (F);

	print "SYSTEM $self->{NAME}: $file.\n"
	    if $dtd->{'VERBOSE'};

	return $_;
    }
}

{
    package SGML::DTDParse::NOTATION;
    @SGML::DTDParse::NOTATION::ISA = qw (SGML::DTDParse::DTD_Element);

    sub parse {
	my($type, $dtd, $notation) = @_;
	my($class) = ref($type) || $type;
	my($self) = {};

	$self->{'DTD'}  = $dtd;
	$self->{'NAME'} = "";
	$self->{'TYPE'} = "";
	$self->{'CONT'} = "";

	if ($notation =~ /<!NOTATION\s+(\S+)\s+(\S+)\s+([\"\'])(.*?)\3\s*>/is) {
	    my($name, $type, $cont) = ($1, $2, $4);
	    $type = uc($type);

	    warn "Unexpected type: $notation\n" 
		if ($type ne 'SYSTEM') && ($type ne 'PUBLIC');

	    $self->{'NAME'} = $name;
	    $self->{'TYPE'} = $type;
	    $self->{'CONT'} = $cont;
	} elsif ($notation =~ /<!NOTATION\s+(\S+)\s+(\S+)\s*>/is) {
	    my($name, $type) = ($1, $2);
	    $type = uc($type);

	    warn "Unexpected type: $notation\n" 
		if ($type ne 'SYSTEM') && ($type ne 'PUBLIC');

	    $self->{'NAME'} = $name;
	    $self->{'TYPE'} = $type;
	} else {
	    print "Notation? $notation\n";
	}

	bless $self, $class;
    }
}

{
    package SGML::DTDParse::ATTLIST;
    @SGML::DTDParse::ATTLIST::ISA = qw (SGML::DTDParse::DTD_Element);

    sub parse {
	my($type, $dtd, $attlist) = @_;
	my($class) = ref($type) || $type;
	my($self) = {};

	$self->{'DTD'}  = $dtd;
	$self->{'NAME'} = "";
	$self->{'CONT'} = "";

	if ($attlist =~ /<!ATTLIST\s+(.*?)\s+(.*?)\s*>\s*$/is) {
	    $self->{'NAME'} = $1;
	    $self->{'CONT'} = $2;
	} else {
	    print "Attlist? $attlist\n";
	}
	
	bless $self, $class;
    }
}

sub new {
    my($type, $dtd_file, $verbose) = @_;
    my($class) = ref($type) || $type;
    my($self) = {};
    my($save_parsed) = "";
    my($done, $e);
    my($entity, $igcount, $element, $attlist, $notation);
    my($parents, $iparents, @dtd, $elemhash, $inclhash);
    my($current);
    local($_, *DTD);

    $self = bless $self, $type;

    $self->{'EXPAND'} = 0;
    $self->{'ATTENT'} = {};
    $self->{'TITLE'} = "";
    $self->{'VERBOSE'} = $verbose;

    $self->{'CATALOG'} = {};
    $self->{'ENTITY'} = {};
    $self->{'ELEMENT'} = {};
    $self->{'NOTATION'} = {};
    $self->{'ATTLIST'} = {};
    
    $self->load_catalogs ();

    open (DTD, $dtd_file) || die "$0: cannot read $dtd_file\n";
    read (DTD, $_, -s $dtd_file);
    close (DTD);

    $_ = $self->expand_dtd ($_);

    s/\n/ /g;         # get rid of newlines
    s/\s+/ /g;        # turn multiple spaces into a single space

    # Find all the element defintions...
    @dtd = m/(<!ELEMENT.*?>)/gi;
    s/<!ELEMENT.*?>//gi;
    
    foreach $element (@dtd) {
	my($elem_param) = "";
	my($pre, $post);

	# Handle <!ELEMENT %foo and <!ELEMENT (foo | bar | baz)
	if ($element =~ /(<!ELEMENT\s+)\((.*?)\)(.*)$/s) {
	    ($pre, $elem_param, $post) = ($1, $2, $3);
	} elsif ($element =~ /(<!ELEMENT\s+)(\%\S+)(.*)$/s) {
	    ($pre, $elem_param, $post) = ($1, $2, $3);
	}

	if ($elem_param ne "") {
	    my(@localelem) = ();
	    my($match) = $elem_param;
	    $match = $self->expand_entities ($match);
	    $match =~ s/[\(\)\s]//g;
	    @localelem = split(/[\|,]/, $match);

	    foreach $element (@localelem) {
		$e = parse SGML::DTDParse::ELEMENT $self, "$pre$element$post";
		$self->define('ELEMENT', $e->name(), $e);
	    }
	} else {
	    $e = parse SGML::DTDParse::ELEMENT $self, $element;
	    $self->define('ELEMENT', $e->name(), $e);
	}
    }

    # Find all the attlist defintions...
    @dtd = m/(<!ATTLIST.*?>)/gi;    
    s/<!ATTLIST.*?>//gi;

    foreach $attlist (@dtd) {
	my($att_param) = "";
	my($pre, $post);

	# Handle <!ATTLIST %foo and <!ATTLIST (foo | bar | baz)
	
	if ($attlist =~ /(<!ATTLIST\s+)\((.*?)\)(.*)$/s) {
	    ($pre, $att_param, $post) = ($1, $2, $3);
	} elsif ($attlist =~ /(<!ATTLIST\s+)(\%\S+)(.*)$/s) {
	    ($pre, $att_param, $post) = ($1, $2, $3);
	}

	if ($att_param ne "") {
	    my(@localatt) = ();
	    my($match) = $att_param;
	    $match = $self->expand_entities ($match);
	    $match =~ s/[\(\)\s]//g;
	    @localatt = split(/[\|,]/, $match);

	    foreach $attlist (@localatt) {
		$e = parse SGML::DTDParse::ATTLIST $self, "$pre$attlist$post";
		$self->define('ATTLIST', $e->name(), $e);
	    }
	} else {
	    $e = parse SGML::DTDParse::ATTLIST $self, $attlist;
	    $self->define('ATTLIST', $e->name(), $e);
	}
    }

    # Find all the notation defintions...
    @dtd = m/(<!NOTATION.*?>)/gi;   
    s/<!NOTATION.*?>//gi;
    foreach $notation (@dtd) {
	$e = parse SGML::DTDParse::NOTATION $self, $notation;
	$self->define('NOTATION', $e->name(), $e);
    }

    # ----------------------------------------

    # Now discard all the <![ INCLUDE [ ... ]]>'s that might be left
    while (/<!\[\s*INCLUDE\s*\[\s*\]\]\>/i) {
	$_ = "$` $'";
    }
    s/\s*//g;
    $self->{'ERROR'} = $_;  # should be "";

    # ----------------------------------------

    # Expand the parameter entities (that aren't files)
    foreach $entity (values %{$self->{'ENTITY'}}) {
	next if !$entity->{'PENT'} || $entity->{'FILE'};
	$entity->{'CONTEX'} = $self->expand_entities($entity->content());
    }

    print "Finding children...\n" if $self->{'VERBOSE'};
    foreach $element (values %{$self->{'ELEMENT'}}) {
	print "Expanding ", $element->name(), "\n" if $self->{'VERBOSE'} > 1;
	$elemhash = {};
	$inclhash = {};

	$_ = $element->{'CONTEX'};
	while (/%([A-Z\-\.0-9]+);?/i) {
	    my($pre, $name, $post) = ($`, $1, $');
	    my($cont) = $self->entity($name)->content();
	    $_ = $pre . $cont . $post;
	}
	s/[^A-Za-z0-9\-\.]/ /g;
	s/^\s*(.*?)\s*$/$1/g;
	foreach $_ (split (/\s+/, $_)) {
	    $elemhash->{uc($_)} = $_;
	}

	$_ = $element->{'INCL'};
	while (/%([A-Z\-\.0-9]+);?/i) {
	    my($pre, $name, $post) = ($`, $1, $');
	    my($cont) = $self->entity($name)->content();
	    $_ = $pre . $cont . $post;
	}
	s/[^A-Za-z0-9\-\.]/ /g;
	s/^\s*(.*?)\s*$/$1/g;
	foreach $_ (split (/\s+/, $_)) {
	    $inclhash->{uc($_)} = $_;
	}

	$_ = $element->{'EXCL'};
	while (/%([A-Z\-\.0-9]+);?/i) {
	    my($pre, $name, $post) = ($`, $1, $');
	    my($cont) = $self->entity($name)->content();
	    $_ = $pre . $cont . $post;
	}
	s/[^A-Z0a-z-9\-\.]/ /g;
	s/^\s*(.*?)\s*$/$1/g;
	foreach $_ (split (/\s+/, $_)) {
	    delete $elemhash->{uc($_)};
	    delete $inclhash->{uc($_)};
	}

	$element->{'CHILDREN'} = $elemhash;
	$element->{'ICHILDREN'} = $inclhash;
    }

    print "Finding parents...\n" if $self->{'VERBOSE'};
    foreach $current (values %{$self->{'ELEMENT'}}) {
	print "Finding parents of ", $current->name(), "\n" 
	    if $self->{'VERBOSE'} > 1;

	$parents = {};
	$iparents = {};

	foreach $element (values %{$self->{'ELEMENT'}}) {
	    $elemhash = $element->{'CHILDREN'};
	    $inclhash = $element->{'ICHILDREN'};

	    if ($inclhash->{uc($current->name())}) {
		$iparents->{$element->name()} = $element->name();
	    } elsif ($elemhash->{uc($current->name())}) {
		$parents->{$element->name()} = $element->name();
	    }
	}

	$current->{'PARENTS'} = $parents;
	$current->{'IPARENTS'} = $iparents;
    }

    return $self;
}

sub expand_dtd {
    my($self) = shift;
    local($_) = shift;
    my($done) = 0;
    my($save_parsed) = "";
    my(@dtd, $entity, $e, $igcount);

    print "Expanding entities...\n" if $self->{'VERBOSE'};
    # Recursively locate and replace parameter entities...

    s/<!--.*?-->//sg;  # discard all comments
    s/--.*?--//sg;     # discard all the comments (in other elements)

    while (/^(.*?)(<!ENTITY.*?>|%[A-Z0-9\.\-]+;?)/is) {
	$save_parsed .= $1;
	$entity = $2;
	$_ = $';
	
#	print "Save: \"$1\"\n";
#	print "Enty: \"$2\"\n";

	if ($entity =~ /<!ENTITY/is) {
	    my($ent1, $ent2) = ($&, $');

	    $entity =~ s/\n/ /g;         # get rid of newlines
	    $entity =~ s/\s+/ /g;        # remove multiple spaces
	    $entity =~ s/<!--.*?-->//g;  # discard all comment decl.
	    $entity =~ s/--.*?--//g;     # discard all the comments elsewhere

	    $e = parse SGML::DTDParse::ENTITY $self, $entity;
	    if ($self->define('ENTITY', $e->name(), $e)) {
		$save_parsed .= $ent1;
		$_ = "$ent2$_";
	    }
	} else {
	    $entity =~ /%([A-Z0-9\.\-]+);?/is;
	    my($name) = $1;
	    my($ent) = $self->entity($name);

	    if ($self->{'VERBOSE'}) {
		# print length($save_parsed), ", ", length($_), ": ";
		print "Substituting for %$name;...\n";
	    }
    
	    die "No entity: $name\n" if !$ent;

	    if ($ent->content_entity()) {
		# It's just some piece of content...leave it in place...
		$save_parsed .= "%" . $name . ";";
	    } else {
		my($content) = $ent->content();

		$_ = $content . $_;

		# What if $content is IGNORE?  Then we better make sure
		# that we aren't now about to enter an IGNORE marked 
		# section...

		if ($content =~ /^ignore$/i) {
		    $save_parsed =~ /^(.*?)(.{0,10})$/s;
		    $save_parsed = $1;
		    $_ = $2 . $_;

		    # Remove IGNORE marked sections...
		    while (/<!\[\s*IGNORE+\s*\[/is) {
			my($pre, $match, $post) = ($`, $&, $');
			my($mc) = new Text::DelimMatch '<!\[.*?\[', '\]\]\>';
			my($ms, $mm, $rest) = $mc->match($match . $post);
			$_ = $pre . $rest;
		    }
		}
	    }
	}
    }

    $_ = $save_parsed . $_;

    return $_;
}

sub load {
    my($type, $file) = @_;
    my($class) = ref($type) || $type;
    my($self) = {};
    my($hash, $e);
    local(*F, $_);

    $self->{'EXPAND'} = 0;
    $self->{'ATTENT'} = {};
    $self->{'TITLE'} = "";

    $self->{'ENTITY'} = {};
    $self->{'ELEMENT'} = {};
    $self->{'NOTATION'} = {};
    $self->{'ATTLIST'} = {};

    $self = bless $self, $class;

    open (F, $file) || die "Cannot open $file.\n";
    while (<F>) {
	chop;
	die "Unexpected object type: $_\n" if !/^SGML\:\:DTDParse\:\:(.*)$/;
	$hash = $1;
	die "Unexpected object type: $_\n"
	    if ($hash !~ /^ENTITY|ELEMENT|NOTATION|ATTLIST$/);

	$e = $_->new($self);
	$e->load(*F);

	$self->{$hash}->{uc($e->name())} = $e;
    }
    close (F);
	    
    return $self;
}

sub store {
    my($self) = shift;
    local(*FILE) = shift;
    my($e);

    print "Storing elements...\n" if $self->{'VERBOSE'};
    foreach $e (values %{$self->{'ELEMENT'}},
		values %{$self->{'ENTITY'}},
		values %{$self->{'ATTLIST'}},
		values %{$self->{'NOTATION'}}) {
	$e->store(*FILE);
    }
}

sub hash_keys {
    my($self, $hash) = @_;

    return sort keys %{$self->{uc($hash)}};
}

sub hash_value {
    my($self, $hash, $key) = @_;

    return $self->{uc($hash)}->{uc($key)};
}

sub hash_set {
    my($self, $hash, $key, $newkey, $value) = @_;
    my($e) = $self->{uc($hash)}->{uc($key)};

    $e->{$newkey} = $value;
}

sub elements {
    my($self) = @_;
    return $self->hash_keys('ELEMENT');
}

sub entities {
    my($self) = @_;
    return $self->hash_keys('ENTITY');
}

sub attlists {
    my($self) = @_;
    return $self->hash_keys('ATTLIST');
}

sub notations {
    my($self) = @_;
    return $self->hash_keys('NOTATION');
}

sub attlist {
    my($self, $name) = @_;
    return $self->hash_value ('ATTLIST', $name);
}

sub entity {
    my($self, $name) = @_;
    return $self->hash_value ('ENTITY', $name);
}

sub element {
    my($self, $name) = @_;
    return $self->hash_value ('ELEMENT', $name);
}

sub notation {
    my($self, $name) = @_;
    return $self->hash_value ('NOTATION', $name);
}

sub expansion_possible {
    my($self) = shift;
    local($_) = shift;

    return 1 if /%([A-Z\-\.0-9]+);?/i;
    return 0;
}

sub expand_one_entity {
    my($self) = shift;
    local($_) = shift;

    if (/%([A-Z\-\.0-9]+);?/i) {
	my($pre, $name, $post) = ($`, $1, $');
	my($ent) = $self->entity($name);
	my($cont) = "\0$name;";

	$cont = $ent->content() if $ent;

	$_ = $pre . $cont . $post;
    }

    return $_;
}

sub expand_entity_cleanup {
    my($self) = shift;
    local($_) = shift;

    s/\0/\%/g;

    return $_;
}

sub expand_entities {
    my($self) = shift;
    local($_) = shift;
    
    while ($self->expansion_possible ($_)) {
	$_ = $self->expand_one_entity($_);
    }

    $_ = $self->expand_entity_cleanup($_);

    return $_;
}

sub load_catalogs {
    my($self) = @_;
    my(@files) = ();
    my($cat) = "";
    my($catpath) = "";
    my(%pubid) = ();
    my($catalog_path) = $SGML::DTDParse::SGML_CATALOG_FILES;
    my($pathsep) = ($^O =~ /mswin/i) ? ";" : ":";
    local($_, *CAT);

    $catalog_path = "./CATALOG$pathsep/usr/local/sgml/CATALOG$pathsep~/sgml/CATALOG"
	if $catalog_path eq "";

    @files = split(/$pathsep/, $catalog_path);
    foreach $cat (@files) {
	print STDERR "Loading catalog: $cat\n" if $self->{'VERBOSE'};
	if ($cat =~ /^(.*)\/[^\/]+$/) {
	    $catpath = $1;
	} else {
	    $catpath = ".";
	}
	open (CAT, $cat) || next;
	while (<CAT>) {
	    chop;
	    
	    if (/^public\s+\"(.*?)\"\s+\"?(\S+)\"?/i) {
		my($pid, $file) = ($1, $2);
		$file = "$catpath/$file" 
		    if ($file !~ /^\//) && ($file !~ /^[a-z]:/i);
		$pubid{$pid} = $file;
		print "$pid\n\t$file\n" if $self->{'VERBOSE'} > 2;
	    } elsif (/^public\s+\"(.*?)\"\s*$/i) {
		my($pid, $file) = ($1, "");
		$_ = scalar(<CAT>);
		s/^\s*\"?(.*?)\"?\s*$/$1/;
		$file = $_;
		$file = "$catpath/$file" 
		    if ($file !~ /^\//) && ($file !~ /^[a-z]:/i);
		$pubid{$pid} = $file;
		print "$pid\n\t$file\n" if $self->{'VERBOSE'} > 2;
	    } else {
		#nop;
	    }
	}
	close (CAT);
    }

    %{$self->{'CATALOG'}} = %pubid;
}

sub define {
    my($self, $type, $name, $value) = @_;

    # In SGML, only the first definition counts, so skip subsequent
    # definitions...

    if (exists $self->{$type}->{uc($name)}) {
#	print "Skipping redefinition of $type $name: ", $value->{'CONT'}, "\n"
#	    if $self->{'VERBOSE'};
	return 0;
    } else {
#	print "Defining $type $name: ", $value->{'CONT'}, "\n"
#	    if $self->{'VERBOSE'};
	$self->{$type}->{uc($name)} = $value;
	return 1;
    }
}

__END__
=head1 NAME

DTDParse - parse a valid SGML DTD and build a simple text database of
the elements, attributes, entities, and notations.

=head1 SYNOPSIS

    use SGML::DTDParse;

    $dtd = new SGML::DTDParse $dtd_file, $verbose

    # Any additional operations you wish to perform...

    open (F, ">$dtd_database");
    $dtd->store(*F);
    close (F);


    $dtd = load SGML::DTParse $dtd_database;

    # Any additional operations you wish to perform...

=head1 DESCRIPTION

DTDParse uses regular expressions to parse the DTD and builds a collection
of objects which represent the elements, attributes, entities, and notations
of the DTD.  Each object is a blessed hash reference.  The members of the
hash are written to the database.  Uppercase names are reserved for DTDParse,
but any values added to the hash before storing the DTD will be written
to the database and reloaded.

=head1 PUBLIC METHODS

Use these methods to build, load, store, and access the database.  There
are a few other methods, used internally, that are not documented here.

=over 4

=item new ($dtd_file, $verbose)

Loads the SGML DTD $dtd_file and builds the database.  The setting of
$verbose determines how much information to print while parsing.
Generally, a higher number means more messages.
Parsing can take
a long time on a large DTD, so it may be valuable to set $verbose to at
least 1.

=item load ($dtd_database)

Loads a DTD from a previously stored database.

=item store (*file)

Stores the current database in *file which should be a file handle glob
currently open for writing.

=item elements ()

Returns a list of the I<names> of the elements in the DTD.

=item entities ()

Returns a list of the I<names> of the entities in the DTD.

=item attlists ()

Returns a list of the I<names> of the attlists in the DTD.

=item notations ()

Returns a list of the I<names> of the notations in the DTD.

=item attribute ($name)

Returns the I<object> associated with the attlist $name.

=item entity ($name)

Returns the I<object> associated with the entity $name.

=item element ($name)

Returns the I<object> associated with the element $name.

=item notation ($name)

Returns the I<object> associated with the notation $name.

=item expand_entities ($string)

Returns $string with all of the parameter entities expanded
(recursively).

=back

=head1 PRIVATE METHODS

These methods aren't really intended for public use, but they may be
valuable for subclasses of DTDParse, so they're documented here.

=over 4

=item hash_set ($type, $name, $member, $value)

This method is a shortcut for setting the value of a member in an
element, entity, attlist, or notation object.  $type should be
'ELEMENT', 'ENTITY', 'ATTLIST', or 'NOTATION'.  $name should be the
name of the element, entity, attlist, or notation.  The member
$member in that object will be set to $value. 

    $dtd->hash_set ('ELEMENT', 'Book', 'hashbase', 'book');

is a shortcut for

    $e = $dtd->element('Book');
    $e->{'hashbase'} = 'book';

=item expansion_possible ($string)

Returns true if $string contains at least one parameter entity.

=item expand_one_entity ($string)

Expands one parameter entity in $string and returns the result.
Note, if you use this method, be sure to call 
C<expand_entity_cleanup ($string)>.
If expand_one_entity finds a parameter entity for which it has no
replacement text, it will replace the C<%> sign of the parameter entity
with an ASCII null.  This prevents expand_one_entity from always attempting
the same macro.  

=item expand_entity_cleanup ($string)

Replaces any ASCII nulls with C<%> signs.  This is cleanup for
C<expand_one_entity>.

=item load_catalogs ()

Reads your C<SGML_CATALOG_FILES> environment variable and does a simple
job of parsing C<PUBLIC> declarations from it for finding parameter
entity replacement text in the DTD.

=item define ($type, $name, $object)

Defines $name of $type ('ELEMENT', 'ENTITY', 'ATTLIST', or 'NOTATION')
to be $object, I<if it is not already defined>.  This follows the SGML
convention that only the first definition of an object counts.

=back

=head1 MEMBER VARIABLES

This section identifies all of the member variables used by
C<SGML::DTDParse>.  Most of these variables contain data that you either
shouldn't mess with or you should get through access methods.  You may
disagree, so here they are:

=head2 SGML::DTDParse MEMBER VARIABLES

=over 4

=item $dtd->{'VERBOSE'}

Stores the verbosity passed to C<parse SGML::DTDParse>.

=item $dtd->{'CATALOG'}

A hash of the public elements parsed from the SGML Open style catalog
by C<load_catalog>.

=item $dtd->{'ENTITY'}

A hash of the entities loaded from the DTD.

=item $dtd->{'ELEMENT'} = {};

A hash of the elements loaded from the DTD.

=item $dtd->{'NOTATION'} = {};

A hash of the elements loaded from the DTD.

=item $dtd->{'ATTLIST'} = {};

A hash of the attlist loaded from the DTD.

=back    

=head2 SGML::DTDParse::* MEMBER VARIABLES

Elements, entities, attribute lists, and notations are kept in objects
of the C<SGML::DTDParse::*> class, where "*" is one of C<ELEMENT>, C<ENTITY>,
C<ATTLIST>, or C<NOTATION>.

Each of these elements has the following member variables:

=over 4

=item $self->{'DTD'}

Contains a pointer to the C<DTD::Parse> object to which this element
(entity, attlist, or notation) belongs.

=item $self->{'NAME'}

Contains the name of this element (entity, attlist, or notation).

=item $self->{'CONT'} = "";

Contains the content of this element (entity, attlist, or notation).

=item $self->{'CONTEX'} = "";

Contains the expanded content of this element (entity, attlist, or notation).

=back

=head2 SGML::DTDParse::ELEMENT MEMBER VARIABLES

In addition, elements have the following member variables:

=over 4

=item $self->{'MIN'}

The allowed minimization of this element.  This is a three character
string, e.g., "- O" for an element with a required start tag and an
optional endt ag.

=item $self->{'INCL'} = "";

The content model fragments that are included in this element.

=item $self->{'EXCL'} = "";

The content model fragments that are excluded in this element.

=back

=head2 SGML::DTDParse::ENTITY MEMBER VARIABLES

In addition, entities have the following member variables:

=over 4

=item $self->{'TYPE'}

The type of entity, none, PUBLIC, or SYSTEM.

=item $self->{'PENT'}

True if this is a parameter entity.

=item $self->{'FILE'}

For SYSTEM or PUBLIC parameter entities, the name of the file that
they point to.

=back

=head2 SGML::DTDParse::ATTLIST MEMBER VARIABLES

Attribute lists have no additional member variables.

=head2 SGML::DTDParse::NOTATION MEMBER VARIABLES

In addition, entities have the following member variable:

=over 4

=item $self->{'TYPE'}

The type of entity, none, PUBLIC, or SYSTEM.

=back

=head1 COPYRIGHT

Copyright (C) 1996 Norman Walsh.  All Rights Reserved.

You are free to use this code in any way you see fit, but give credit
where it's due.

=cut
