# -*- Perl -*-

package SGML::DTDParse::Html;

use strict;
require 5.000;
require Carp;
require 'dtdlib/base.pl';

$SGML::DTDParse::Html::VERSION = $SGML::DTDParse::VERSION;

@SGML::DTDParse::Html::ISA = qw (SGML::DTDParse);

{
    package SGML::DTDParse::DTD_Element;
    # Extension to package defined in DTD.pm

    sub html_file {
	my($self, $hash, $name) = @_;
	my($html) = "";

	# Usage: ->html_file('element', $name) # for element $name
	#     or ->html_file('entity', $name)  # for entity $name
	#     or ->html_file()                 # for this name

	if (!defined ($hash) || !defined ($name)) {
	    $html = &main::Filename(ref $self, $self->name());
	} else {
	    $html = &main::Filename($hash, $name);
	}

	$html .= ".htm";
	return $html;
    }

    sub print_content_model {
	my($self) = shift;
	local($_) = shift;
	local(*FILE) = shift;

	my(@cm) = $self->format_content_model($_);
	foreach $_ (@cm) {
	    print FILE $self->make_links($_, 1, 1), "\n";
	}
    }

    sub make_links {
	my($self) = shift;
	local($_) = shift;
	my($elements, $entities) = @_;
	my($dtd) = $self->{'DTD'};
	my($out) = "";

	while ($_ =~ /^(.)(.*)$/) {
	    my($first, $rest) = ($1, $2);

#	    print "f: $first\n";
#	    print "r: $rest\n";

	    if (($first eq "%") && $entities) {
		my($match, $name, $url);
		my($dir) = $dtd->entity_dir();

		($match = $rest) =~ /^([A-Z0-9\-\.]+);?/i;
		$name = $1;
		$url  = $1;
		$rest = $';

		$url = $self->html_file('entity', $url);
		$url = "../$dir/$url";

		if ($self->{'DTD'}->entity($name)) {
		    $out .= "%<A HREF=\"$url\">$name</A>;";
		} else {
		    $out .= "%$name;";
		}
	    } elsif (($first =~ /[A-Z0-9\-\.]/i) && $elements) {
		my($match, $one, $url, $name);
		my($dir) = $self->{'DTD'}->element_dir();

		if ($rest =~ /^([A-Z0-9\-\.]+)/i) {
		    $match = $1;
		    $name = "$first$1";
		    $url  = "$first$1";
		    $rest = $';

#		    print "Rest: $rest\n";
#		    print "Mtch: $&\n";
#		    print "left: $'\n\n";

		} else {
		    $match = "";
		    $name = $first;
		    $url = $first;
		}

		$url = $self->html_file('element', $url);
		$url = "../$dir/$url";

		if ($self->{'DTD'}->element($name)) {
		    $out .= "<A HREF=\"$url\">$name</A>";
		} else {
		    $out .= $name;
		}
	    } else {
		$out .= $first;
	    }

#	    print "F: $first\n";
#	    print "R: $rest\n";
	    
	    $_ = $rest;

#	    print "_: $_\n";
	}

#	print "L: $_\n";
	
	$out .= $_;
	return $out;
    }	

    sub parse_attributes {
	my($self) = shift;
	local($_) = shift;
	my($dtd) = $self->{'DTD'};
	my($pre, $att, $rest, @attr, @exattr);
	my($skip, $skiptext);

	$_ = $dtd->expand_entities($_); # if $dtd->{'EXPAND'}; 

	@attr = ();
	s/\s+/ /g;
	while ($_ !~ /^\s*$/) {
	    my($mc) = new Text::DelimMatch;
	    s/^\s*(.*?)\s*$/$1/g;

	    if (/^\(/) {
		$mc->delim('\(', '\)');
		($pre, $att, $_) = $mc->match ($_);
	    } elsif (/^\'/) {
		$mc->delim("'");
		($pre, $att, $_) = $mc->match ($_);
	    } elsif (/^\"/) {
		$mc->delim('"');
		($pre, $att, $_) = $mc->match ($_);
	    } else {
		/^(\S+)\s*(.*)$/;
		$pre = "";
		$att = $1;
		$_ = $2;
	    }

	    push (@attr, "$pre$att");
	}

	@exattr = ();
	while (@attr) {
	    $att = shift @attr;
	    if ($att =~ /^\%/) {
		push (@exattr, ($att, "", "", ""));
	    } else {
		my($values) = shift @attr;
		my($defval) = shift @attr;
		my($type) = "&nbsp;";

		if (($defval) =~ /^\#fixed/i) {
		    $type = $defval;
		    $defval = shift @attr;
		} elsif ($defval =~ /^\#/i) {
		    $type = $defval;
		    $defval = "&nbsp;";
		}

		if ($values =~ /NOTATION/i) {
		    $type = $values;
		    $values = $defval;
		    $defval = shift @attr;
		}

		$values =~ s/\|/ \| /g; # allow long lists of attribute
		                        # values to break properly in tables

		push (@exattr, ($att, $values, $type, $defval));
	    }
	}

	return @exattr;
    }

    sub format_attributes {
	my($self) = shift;
	local($_) = shift;
	local(*HTML) = shift;
	my($dtd) = $self->{'DTD'};
	my($para) = "";
	my($att, $rest, @attr);
	my($skip, $skiptext);

	@attr = $self->parse_attributes ($_);

	return if !@attr;

	($para, @attr) = &main::analyze_attributes($dtd,@attr);

	if ($para) {
	    print HTML "<P>\n";
	    print HTML "$para\n";
	}

	return if !@attr;

	print HTML "<P>\n";
	print HTML "<TABLE BORDER=1>\n";
	print HTML "<TR>\n";
	print HTML "<TH ALIGN=LEFT VALIGN=TOP>Name</TH>\n";
	print HTML "<TH ALIGN=LEFT VALIGN=TOP>Value(s)</TH>\n";
	print HTML "<TH ALIGN=LEFT VALIGN=TOP>Type</TH>\n";
	print HTML "<TH ALIGN=LEFT VALIGN=TOP>Default Value</TH>\n";
	print HTML "</TR>\n";

	while (@attr) {
	    $att = shift @attr;
	    if ($att =~ /^\%/) {
		splice (@attr, 0, 3);
		print HTML "<TR>\n";
		print HTML "<TD ALIGN=LEFT VALIGN=TOP>", $self->make_links($att, 0, 1), "</TD>\n";
		print HTML "<TD COLSPAN=3 ALIGN=LEFT VALIGN=TOP>&nbsp;</TD>\n";
		print HTML "</TR>\n";
	    } else {
		my ($values, $type, $defval) = splice (@attr, 0, 3);

		if ($main::ATTR_HTML{uc($att)}) {
		    $att = sprintf ("<A HREF=\"../%s/%s.htm\">%s</A>",
				    $dtd->attribute_dir(),
				    $main::ATTR_HTML{uc($att)},
				    $att);
		}

		$values = $self->make_links($values, 0, 1);
		$type = $self->make_links($type, 0, 1);
		$defval = $self->make_links($defval, 0, 1);

		print HTML "<TR>\n";
		print HTML "<TD ALIGN=LEFT VALIGN=TOP>$att</TD>\n";
		print HTML "<TD ALIGN=LEFT VALIGN=TOP>$values</TD>\n";
		print HTML "<TD ALIGN=LEFT VALIGN=TOP>$type</TD>\n";
		print HTML "<TD ALIGN=LEFT VALIGN=TOP>$defval</TD>\n";
		print HTML "</TR>\n";
	    }
	}
	
	print HTML "</TABLE>\n\n";
    }
}

{
    package SGML::DTDParse::ELEMENT;
    # Extension to package defined in DTD.pm

    sub html {
	my($self) = shift;
	local(*HTML) = shift;
	my($name) = $self->name();
	my($smin, $emin, $incl, $excl, $cont) = ("", "", "", "", "");
	my($file) = $name;
	my (%min) = ('--' => "Both the start- and end-tags are required for this element.",
		     'OO' => "Both the start- and end-tags are optional for this element, if your SGML declaration allows tag minimization.",
		     'O-' => "The start-tag is optional for this element, if your SGML declaration allows tag minimization.  The end-tag is required.",
		     '-O' => "The start-tag is required for this element.  The end-tag is optional, if your SGML declaration allows minimization."
		    );
	my(@children, @parents, $url);
	my(@ichildren, @iparents, $num_children);
	my($expansion, $refpurpose, @pe, $entity, $pos);
	my($description, $attrdesc, $example);
	local($_);

	if ($self->{'MIN'} =~ /^(.) (.)/) {
	    $smin = $1;
	    $emin = $2;
	} else {
	    $smin = "-";
	    $emin = "-";
	}

	$incl = $self->{'INCL'};
	$excl = $self->{'EXCL'};
	
#	print HTML "<HTML>\n<HEAD>\n<TITLE>";
#	print HTML $self->{'DTD'}->{'TITLE'}, " Element $name</TITLE>\n";
#	print HTML "</HEAD><BODY BGCOLOR=\"#FFFFFF\">\n";

	if ($self->{'DTD'}->{'EXPAND'}) {
	    my($url) = $self->html_file('element', $name);
	    $url = "../" . $self->{'DTD'}->element_dir(0) . "/$url";
	    print HTML "<H1>Element $name</H1>\n";
	    $cont = $self->{'CONTEX'};
	    $expansion = "";
	} else {
	    my($url) = $self->html_file('element', $name);
	    $url = "../" . $self->{'DTD'}->element_dir(1) . "/$url";
	    print HTML "<H1>Element $name</H1>\n";
	    $cont = $self->{'CONT'};
	    $expansion = "(unexpanded)";
	}

	$refpurpose = &main::RefPurpose($self);

	if ($refpurpose) {
	    print HTML $refpurpose;
	    print HTML "\n";
	}

	print HTML "<H2>Synopsis</H2>\n\n";

	print HTML "<H3>Content Model</H3>\n";
	print HTML "<P>\n";

	print HTML "<PRE>\n";
	$self->print_content_model ($cont, *HTML);
	print HTML "</PRE>\n";

	if ($incl) {
	    print HTML "<H3>Inclusions</H3>\n";
	    print HTML "<P>\n";
	    print HTML "$name includes these elements at every level:\n";
	    print HTML "<P>\n";
	    print HTML "<PRE>\n";
	    $self->print_content_model ($incl, *HTML);
	    print HTML "</PRE>\n\n";
	}

	if ($excl) {
	    print HTML "<H3>Exclusions</H3>\n";
	    print HTML "<P>\n";
	    print HTML "$name excludes these elements at every level:\n";
	    print HTML "<P>\n";
	    print HTML "<PRE>\n";
	    $self->print_content_model ($excl, *HTML);
	    print HTML "</PRE>\n\n";
	}

	if ($self->{'DTD'}->attlist($self->name())) {
	    my($attr) = $self->{'DTD'}->attlist($self->name());

	    print HTML "<H3>Attributes</H3>\n";

	    $attr->html(*HTML);
	}

	print HTML "<H3>Tag Minimization</H3>\n";
	print HTML "<P>\n";
	print HTML $min{$smin . $emin}, "\n";

	@pe = ();
	foreach $entity (values %{$self->{'DTD'}->{'ENTITY'}}) {
	    my($cont) = $entity->{'CONTEX'};
	    next if !$entity->{'PENT'} || $entity->{'FILE'};
	    $cont = " $cont ";
	    $cont =~ s/[\|\&\,]/ /g;
	    push (@pe, $entity) if $cont =~ / $name /i;
	}

	if (!$self->{'DTD'}->{'EXPAND'} && @pe) {
	    print HTML "<H3>Parameter Entities</H3>\n";
	    print HTML "<P>\n";
	    print HTML "$name appears in the following parameter entities:\n";

	    while (@pe) { 
		my($entity) = shift @pe;
		my($dir) = $self->{'DTD'}->entity_dir();
		$url = $self->html_file('entity', $entity->name());
		$url = "%<A HREF=\"../$dir/$url\">" . $entity->name() . "</A>;";

		$url .= ", " if @pe;
		$pos += length($url);
		print HTML $url;
		if ($pos > 70) {
		    $pos = 0;
		    print HTML "\n";
		}
	    }
	    print HTML "\n\n";
	}

	@parents = sort split(/\s+/, $self->{'PARENTS'});
	@iparents = sort split(/\s+/, $self->{'IPARENTS'});

	@children = split(/\s+/, $self->{'CHILDREN'});
	@ichildren = split(/\s+/, $self->{'ICHILDREN'});
	push (@children, @ichildren);
	@children = sort @children;
	$num_children = $#children+1;

	$incl = 0;
	$excl = 0;
	foreach $_ (@parents) {
	    my($elem) = $self->{'DTD'}->element($_);
	    $incl = 1 if ($elem && $elem->{'INCL'});
	    $excl = 1 if ($elem && $elem->{'EXCL'});
	    last if $incl && $excl;
	}

	if ($children[0] =~ /^empty$/i) {
	    # An empty element cannot meaningfully have inclusions or
	    # exclusions...
	    $incl = 0;
	    $excl = 0;
	}

	$description = &main::Description($self);

	if ($description || @parents || $incl || @children) {
	    print HTML "<H2>Description</H2>\n\n";
	}

	if ($description) {
	    print HTML $description;
	    print HTML "\n";
	}

	if (@parents) {
	    my($comma) = "";
	    $pos = 0;
	    print HTML "<H3>Parents</H3>\n";
	    print HTML "<P>\n";

	    foreach $_ (@parents) {
		$url = $self->html_file('element', $_);

		if ($self->{'DTD'}->element($_)) {
		    my($dir) = $self->{'DTD'}->element_dir();
		    $url = "<A HREF=\"../$dir/$url\">$_</A>";
		}

		$pos += length($url) + length($comma);
		print HTML $comma;
		if ($pos > 70) {
		    $pos = length($url);
		    print HTML "\n";
		}
		print HTML $url;
		
		$comma = ", ";
	    }

	    print HTML "\n\n";
	}

	if (@iparents) {
	    my($comma) = "";
	    $pos = 0;
	    print HTML "<P>\n";
	    print HTML "These elements contain $name at all levels by inclusion.\n";
	    print HTML "<P>\n";

	    foreach $_ (@iparents) {
		$url = $self->html_file('element', $_);

		if ($self->{'DTD'}->element($_)) {
		    my($dir) = $self->{'DTD'}->element_dir();
		    $url = "<A HREF=\"../$dir/$url\">$_</A>";
		}

		$pos += length($url) + length($comma);
		print HTML $comma;
		print HTML $url;
		if ($pos > 70) {
		    $pos = 0;
		    print HTML "\n";
		}
		
		$comma = ", ";
	    }

	    print HTML "\n\n";
	}

	if ($incl || @children) {
	    $pos = 0;
	    print HTML "<H3>Children</H3>\n";
	    print HTML "<P>\n";

	    while (@children) { 
		$_ = shift @children;

		if ($self->{'DTD'}->element($_)) {
		    my($dir) = $self->{'DTD'}->element_dir();
		    $url = $self->html_file('element', $_);
		    $url = "<A HREF=\"../$dir/$url\">$_</A>";
		} else {
		    $url = $_;
		}

		$url .= ", " if @children;
		$pos += length($url);
		print HTML $url;
		if ($pos > 70) {
		    $pos = 0;
		    print HTML "\n";
		}
	    }

	    if ($incl || $excl) {
		if ($num_children > 0) {
		    print HTML "\n\n";
		    print HTML "<P>\n";
		}

		if ($incl) {
		    print HTML "In some contexts, additional elements ";
		    print HTML "may be allowed\n";
		    print HTML "by inclusions in parent elements.\n";
		} 

		if ($excl) {
		    if ($incl) {
			print HTML "Similarly, in ";
		    } else {
			print HTML "In ";
		    }

		    print HTML "some contexts, some elements\n";
		    print HTML "may be invalid ";
		    print HTML "due to exclusions in parent elements.\n";
		}
		print HTML "\n";
	    } else {
		print HTML "\n\n";
	    }
	}

	if ($self->{'DTD'}->attlist($self->name())) {
	    my($attr) = $self->{'DTD'}->attlist($self->name());
	    $attrdesc = &main::AttributeDescription($attr);

	    if ($attrdesc) {
		print HTML "<H2>Attributes</H2>\n";
		print HTML "\n$attrdesc\n\n";
	    }
	}

	$example = &main::Example($self);

	if ($example) {
	    print HTML "<H2>Examples</H2>\n";
	    print HTML "$example\n\n";
	}

#	print HTML "\n</BODY>\n</HTML>\n";
#	close (HTML);
    }
}

{
    package SGML::DTDParse::ENTITY;
    # Extension to package defined in DTD.pm

    sub html {
	my($self) = shift;
	local(*HTML) = shift;
	my($name) = $self->name();
	my($type, $cont, $pent, $file) = ("", "", "", "");
	my($html) = $name;
	my($expansion, $refpurpose, @pe, $entity, $url, $pos);
	my($description);
	local($_);

	$html =~ tr/A-Z/a-z/;
	$html .= ".html";

	$type = $self->{'TYPE'};
	$cont = $self->{'CONT'};
	$pent = $self->{'PENT'};
	$file = $self->{'FILE'};

#	print HTML "<HTML>\n<HEAD>\n<TITLE>";
#	print HTML $self->{'DTD'}->{'TITLE'}, " Entity $name</TITLE>\n";
#	print HTML "</HEAD><BODY BGCOLOR=\"#FFFFFF\">\n";

	if ($self->{'DTD'}->{'EXPAND'}) {
	    my($url) = $self->html_file('entity', $name);
	    $url = "../" . $self->{'DTD'}->entity_dir(0) . "/$url";
	    print HTML "<H1>Entity $name</H1>\n";
	    $expansion = "";
	} else {
	    my($url) = $self->html_file('entity', $name);
	    $url = "../" . $self->{'DTD'}->entity_dir(1) . "/$url";
	    print HTML "<H1>Entity $name</H1>\n";
	    $expansion = "(unexpanded)";
	}

	$refpurpose = &main::RefPurpose($self);

	if ($refpurpose) {
	    print HTML $refpurpose;
	    print HTML "\n";
	}

	print HTML "\n<H3>Expansion</H3>\n";
	print HTML "<P>\n";

	# see if we can figure out that it's an attribute...
	if ($name =~ /attrib$/
	    || $cont =~ /\#IMPLIED/i
	    || $cont =~ /\#REQUIRED/i) {
	    $self->format_attributes ($cont, *HTML);
	} else {
	    print HTML "<PRE>\n";
	    $self->print_content_model ($cont, *HTML);
	    print HTML "</PRE>\n";
	}

	@pe = ();
	foreach $entity (values %{$self->{'DTD'}->{'ENTITY'}}) {
	    my($cont) = $entity->{'CONTEX'};
	    next if !$entity->{'PENT'} || $entity->{'FILE'};
	    $cont = " $cont ";
	    $cont =~ s/[\|\&\,]/ /g;
	    push (@pe, $entity) if $cont =~ / %$name; /i;
	}

	if (!$self->{'DTD'}->{'EXPAND'} && @pe) {
	    print HTML "<H3>Parameter Entities</H3>\n";
	    print HTML "<P>\n";
	    print HTML "%$name; appears in the following parameter entities:\n";

	    while (@pe) { 
		my($entity) = shift @pe;
		my($dir) = $self->{'DTD'}->entity_dir();
		$url = $self->html_file('entity', $entity->name());
		$url = "%<A HREF=\"../$dir/$url\">" . $entity->name() . "</A>;";

		$url .= ", " if @pe;
		$pos += length($url);
		print HTML $url;
		if ($pos > 70) {
		    $pos = 0;
		    print HTML "\n";
		}
	    }
	    print HTML "\n\n";
	}

	$description = &main::Description($self);

	if ($description && $self->{'DTD'}->{'EXPAND'}) {
	    print HTML "<H3>Description</H3>\n\n";
	    print HTML $description;
	    print HTML "\n";
	}

#	print HTML "\n</BODY>\n</HTML>\n";
    }
}

{
    package SGML::DTDParse::NOTATION;
    # Extension to package defined in DTD.pm

    sub html {
	my($self) = @_;
	return;
    }
}

{
    package SGML::DTDParse::ATTLIST;
    # Extension to package defined in DTD.pm

    sub html {
	my($self) = shift;
	local(*HTML) = shift;
	my($name) = $self->{'NAME'};
	my($att, $rest, @attr);
	local($_);

	$_ = $self->content();
	return if !defined($_);

	$self->format_attributes($_, *HTML);
    }
}

sub load {
    my($type) = shift;
    my($class) = ref($type) || $type;
    my($self) = load SGML::DTDParse @_;
    return bless $self, $type;
}

sub html {
    my($self) = shift;
    my($element) = shift;
    my($name) = shift;
    local(*FILE) = shift;

    if (! exists $self->{$element}->{uc($name)}) {
	warn "There is no $element \"$name\".\n";
	return;
    }

    $self->{$element}->{uc($name)}->html(*FILE, @_);
}

sub entity_dir {
    my($self, $expand) = @_;
    my($entities) = $self->{'EXPANDED_ENTITY_DIR'} || "entities";
    my($dtdent)   = $self->{'UNEXPANDED_ENTITY_DIR'} || "dtdent";

    $expand = $self->{'EXPAND'} if !defined($expand);
    
    return $expand ? $entities : $dtdent;
}	

sub element_dir {
    my($self, $expand) = @_;
    my($elements) = $self->{'EXPANDED_ELEMENT_DIR'} || "elements";
    my($dtdelem)  = $self->{'UNEXPANDED_ELEMENT_DIR'} || "dtdelem";

    $expand = $self->{'EXPAND'} if !defined($expand);

    return $expand ? $elements : $dtdelem;
}	

sub attribute_dir {
    my($self, $expand) = @_;
    my($attr) = $self->{'EXPANDED_ATTRIBUTE_DIR'} || "attr";
    my($dtdattr)  = $self->{'UNEXPANDED_ATTRIBUTE_DIR'} || "dtdattr";

    $expand = $self->{'EXPAND'} if !defined($expand);

    return $expand ? $attr : $dtdattr;
}	

1;

__END__
=head1 NAME

SGML::DTDParse::Html - produce a set of HTML web pages for a DTD.

=head1 SYNOPSIS

SGML::DTDParse::Html is a subclass of SGML::DTDParse

    use SGML::DTDParse::Html;

    $dtd = load SGML::DTDParse::Html $dtd_database;

    foreach $name ($dtd->elements()) {
       my($element)  = $dtd->element($name);
       my($htmlfile) = $element->html_file();

       print "Writing $name...\n";

       open (HTML, ">$file");
       $element->html(*HTML);
       close (HTML);
    }

    # Similarly for entities and notations.

=head1 DESCRIPTION

DTDParse::Html prints out an HTML page for each element and entity.
Attribute lists are automatically presented in the element.  Notations
aren't currently handled, mostly because it isn't clear to me how
they would be usefully presented.

=head1 PUBLIC METHODS

There are no particularly useful public methods of C<SGML::DTDParse::Html>

=head1 MEMBER VARIABLES

These variables control some aspects of the translation.  They really should
have access methods, but I haven't bothered to code them up.

=over 4

=item $dtd->{'EXPAND'}

Controls expansion of parameter entities.  If true, all parameter entities
are expanded, but see C<ATTLIST> below.

=item $dtd->{'TITLE'}

The title for the DTD.  Usually something like "DTDName ver#" is sufficient
(e.g., "DocBook 3.0").

=item $dtd->{'SHORTTITLE'}

A short title for the DTD.  Usually just the DTD name.

=item $dtd->{'EXPANDED_ENTITY_DIR'}

=item $dtd->{'UNEXPANDED_ENTITY_DIR'}

=item $dtd->{'EXPANDED_ELEMENT_DIR'}

=item $dtd->{'UNEXPANDED_ELEMENT_DIR'}

=item $dtd->{'EXPANDED_TREE_DIR'}

=item $dtd->{'UNEXPANDED_TREE_DIR'}

When constructing links to other HTML pages, C<SGML::DTDParse::Html> can
determine the name of the file that will be used for that page, but cannot
know what directories they will be stored in.  It assumes that the
directories specified in these member variables are correct.  The expanded
versions are used if C<$dtd->{'EXPAND'}> is true, otherwise the unexpanded
versions are used.

=back

=head2 Callbacks

These objects use the callbacks described in the documentation for
dtd2html to incorporate additional documentation into the HTML pages.

=head1 COPYRIGHT

Copyright (C) 1996 Norman Walsh.  All Rights Reserved.

You are free to use this code in any way you see fit, but give credit
where it's due.

=cut
